package edu.princeton.swing;

import javax.swing.*;

/**
 * DefaultTristateButtonModel is a subclass of DefaultButtonModel to implement TristateButtonModel.
 *
 * @author btsang
 * @version 7.1
 */
public class DefaultTristateButtonModel extends JToggleButton.ToggleButtonModel
                                        implements TristateButtonModel
{
    private boolean tristate;

    /**
     * Constructs a DefaultTristateButtonModel.
     */
    public DefaultTristateButtonModel()
    {
        super();

        tristate = false;
    }

    /**
     * Constructs a DefaultTristateButtonModel.
     */
    public DefaultTristateButtonModel(ButtonModel model)
    {
        super();

        stateMask = 
            (model.isArmed()    ? ARMED    : 0) |
            (model.isEnabled()  ? ENABLED  : 0) |
            (model.isPressed()  ? PRESSED  : 0) |
            (model.isRollover() ? ROLLOVER : 0) |
            (model.isSelected() ? SELECTED : 0);
        actionCommand = model.getActionCommand();
        mnemonic = model.getMnemonic();

        tristate = false;
    }

    /**
     * Returns true if the button is in tristate.  If this returns true, isSelected() should also
     * return true.
     *
     * @return True iff the button is in tristate.
     */
    public boolean isTristate()
    {
        return tristate;
    }

    /**
     * Sets wheter or not the button is in tristate.  If true is passed, setSelected(true)
     * would also be called.
     *
     * @param tristate Wheter or not the button should be in tristate.
     */
    public void setTristate(boolean tristate)
    {
        if (tristate == this.tristate)
            return;

        this.tristate = tristate;

        if (tristate && !isSelected())
        {
            super.setSelected(true);
        }
        else {
            fireStateChanged();
        }
    }
        
    /**
     * Selects or deselects the button.  This method is overridden to take the button out of
     * tristate.
     */
    public void setSelected(boolean b)
    {
        if (!tristate && isSelected() == b)
            return;

        tristate = false;

        if (isSelected() != b)
        {
            super.setSelected(b);
        }
        else {
            fireStateChanged();
        }
    }
}
