package edu.princeton.toy;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Iterator;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.tree.*;

import edu.princeton.swing.*;
import edu.princeton.swing.text.*;
import edu.princeton.toy.lang.*;
import edu.princeton.toy.choosers.*;

public class TOptionsFrame extends JFrame implements ActionListener, TreeSelectionListener
{
    public static final int WIDTH = 640;
    public static final int HEIGHT = 460;

    private static final String CLASS_STRING = TOptionsFrame.class.toString();

    public static final String       OK_COMMAND = CLASS_STRING + "#okCommand";
    public static final String   CANCEL_COMMAND = CLASS_STRING + "#cancelCommand";
    public static final String DEFAULTS_COMMAND = CLASS_STRING + "#defaultsCommand";
    public static final String    APPLY_COMMAND = CLASS_STRING + "#applyCommand";

    public static final DefaultMutableTreeNode ROOT_NODE =
        new DefaultMutableTreeNode("Visual X-TOY Options", true);

    public static final DefaultMutableTreeNode EDITING_NODE =
        new DefaultMutableTreeNode("Editing", true);
    public static final DefaultMutableTreeNode EDITING_FONTS_NODE =
        new DefaultMutableTreeNode("Fonts", false);
    public static final DefaultMutableTreeNode EDITING_COLORS_NODE =
        new DefaultMutableTreeNode("Colors", false);
    public static final DefaultMutableTreeNode EDITING_HIGHLIGHTING_NODE =
        new DefaultMutableTreeNode("Syntax Highlighting", false);
    public static final DefaultMutableTreeNode EDITING_AUTO_COMPLETE_NODE =
        new DefaultMutableTreeNode("Auto-Complete", false);
    public static final DefaultMutableTreeNode EDITING_CHECKING_NODE =
        new DefaultMutableTreeNode("Syntax Checking", false);

    public static final DefaultMutableTreeNode EXECUTION_NODE =
        new DefaultMutableTreeNode("Execution", true);
    public static final DefaultMutableTreeNode EXECUTION_EXCEPTION_NODE =
        new DefaultMutableTreeNode("Exception Handling", false);
    public static final DefaultMutableTreeNode EXECUTION_PERFORMANCE_NODE =
        new DefaultMutableTreeNode("Performance", false);

    public static final String FONT_AREA_NAMES[] =
    {
        "Program Text Area",
        "Warning List",
        "Reference Panel",
        "Stdin Panel",
        "Stdout Panel",
        "Core Panel"
    };

    public static final String FONT_AREA_KEYS[] =
    {
        "toy.editing.fonts#programTextAreaFont",
        "toy.editing.fonts#warningListFont",
        "toy.editing.fonts#referencePanelFont",
        "toy.editing.fonts#stdinPanelFont",
        "toy.editing.fonts#stdoutPanelFont",
        "toy.editing.fonts#corePanelFont"
    };

    public static final int PROGRAM_AREA_INDEX = 0;
    public static final int WARNING_AREA_INDEX = 1;
    public static final int REFERENCE_AREA_INDEX = 2;
    public static final int STDIN_AREA_INDEX = 3;
    public static final int STDOUT_AREA_INDEX = 4;
    public static final int CORE_AREA_INDEX = 5;

    public static final String COLOR_NAMES[] =
    {
        "Foreground",
        "Background",
        "Caret",
        "Selection",
        "Selected Text",
        "Highlighted Line",
        "Highlighted Text",
        "Line Numbers",
        "Column Markers"
    };

    public static final String COLOR_DISABLED_DESCRIPTIONS[] =
    {
        "Use default",
        "Use default",
        "Use foreground",
        "Don't paint",
        "Same as normal text",
        "Don't paint",
        "Same as normal text",
        "Don't show",
        "Don't paint"
    };

    public static final String COLOR_COLOR_KEYS[] =
    {
        "toy.editing.colors.foreground#color",
        "toy.editing.colors.background#color",
        "toy.editing.colors.caret#color",
        "toy.editing.colors.selection#color",
        "toy.editing.colors.selectedText#color",
        "toy.editing.colors.highlightedLine#color",
        "toy.editing.colors.highlightedText#color",
        "toy.editing.colors.lineNumbers#color",
        "toy.editing.colors.columnMarkers#color"
    };

    public static final String COLOR_DISABLED_KEYS[] =
    {
        "toy.editing.colors.foreground#disabled",
        "toy.editing.colors.background#disabled",
        "toy.editing.colors.caret#disabled",
        "toy.editing.colors.selection#disabled",
        "toy.editing.colors.selectedText#disabled",
        "toy.editing.colors.highlightedLine#disabled",
        "toy.editing.colors.highlightedText#disabled",
        "toy.editing.colors.lineNumbers#disabled",
        "toy.editing.colors.columnMarkers#disabled"
    };

    public static final int FOREGROUND_COLOR_INDEX = 0;
    public static final int BACKGROUND_COLOR_INDEX = 1;
    public static final int CARET_COLOR_INDEX = 2;
    public static final int SELECTION_COLOR_INDEX = 3;
    public static final int SELECTED_TEXT_COLOR_INDEX = 4;
    public static final int HIGHLIGHTED_LINE_COLOR_INDEX = 5;
    public static final int HIGHLIGHTED_TEXT_COLOR_INDEX = 6;
    public static final int LINE_NUMBERS_COLOR_INDEX = 7;
    public static final int COLUMN_MARKERS_COLOR_INDEX = 8;


    /**
     * A description of the syntax styles used by a TProgramDocument.
     */
    public static final String HIGHLIGHT_STYLE_NAMES[] =
    {
        "Identifier",
        "Instruction",
        "Keyword",
        "Comment",
        "Auto-Comment",
        "Outside Margin"
    };

    /**
     * The keys corresponding to HIGHLIGHT_STYLE_NAMES.
     */
    public static final String HIGHLIGHT_STYLE_KEYS[] =
    {
        "toy.editing.highlighting#identifier",
        "toy.editing.highlighting#instruction",
        "toy.editing.highlighting#keyword",
        "toy.editing.highlighting#comment",
        "toy.editing.highlighting#autoComment",
        "toy.editing.highlighting#outsideMargin"
    };

    /**
     * A set of AutoCompleters which can be used by a PHighlightedTextArea.
     */
    public static final AutoCompleter AUTO_COMPLETERS[] =
    {
        new IndentingAutoCompleter(),
        new TUncommentingAutoCompleter(),
        new TCommentingAutoCompleter()
    };

    /**
     * The names for the AutoCompleters in the AUTO_COMPLETERS array.
     */
    public static final String AUTO_COMPLETER_NAMES[] =
    {
        "No Auto-completion",
        "Uncommenting Auto-completion",
        "Commenting and Uncommenting Auto-completion"
    };

    /**
     * The key for the autocompleter entry.
     */
    public static final String AUTO_COMPLETER_KEY = "toy.editing.autoComplete#autoCompleter";

    /**
     * The key for wheter or not to ignore warnings.
     */
    public static final String IGNORE_WARNINGS_KEY = "toy.editing.checking#ignoreWarnings";

    /**
     * The key for wheter or not to automatically check the user's syntax as he types.
     */
    public static final String AUTO_CHECK_KEY = "toy.editing.checking#autoCheck";

    /**
     * The keys corresponding to TExceptionType.TYPES.
     *
     * @see edu.princeton.toy.lang.TExceptionType#TYPES
     */
    public static final String EXCEPTION_TYPE_KEYS[] =
    {
        "toy.execution.exceptionHandling#willThrowRegisterUninitialized",
        "toy.execution.exceptionHandling#willThrowMemoryUninitialized",
        "toy.execution.exceptionHandling#willThrowCommandUninitialized",
        "toy.execution.exceptionHandling#willThrowOverflow",
        "toy.execution.exceptionHandling#willThrowShiftOutOfBounds",
        "toy.execution.exceptionHandling#willThrowPcOutOfBounds",
        "toy.execution.exceptionHandling#willThrowRegisterOutOfBounds",
        "toy.execution.exceptionHandling#willThrowMemOutOfBounds"
    };

    /**
     * The key for the clock periods.
     */
    public static final String CLOCK_PERIOD_KEY = "toy.execution.performance#clockPeriod";

    /**
     * The possible values for the clock periods.
     */
    public static final int CLOCK_PERIODS[] =
    {
        0,    1,    2,    3,    4,    5,    6,    7,    8,    9,
        10,   12,   16,   20,   25,   32,   40,   50,   63,   80,
        100,  120,  160,  200,  250,  320,  400,  500,  630,  800,
        1000, 1200, 1600, 2000, 2500, 3200, 4000, 5000, 6300, 8000,
        10000
    };

    /**
     * The key for the clock periods.
     */
    public static final String REFRESH_RATE_KEY = "toy.execution.performance#refreshRate";

    /**
     * The possible refresh rate values.
     */
    public static final int REFRESH_RATES[] =
    {
              1,    2,    3,    4,    5,    6,    7,    8,    9,
        10,   13,   16,   20,   25,   32,   40,   50,   63,   79,
        100,  130,  160,  200,  250,  320,  400,  500,  630,  790,
        1000, 1300, 1600, 2000, 2500, 3200, 4000, 5000, 6300, 7900,
        10000
    };

    private static final Toolkit TOOLKIT = Toolkit.getDefaultToolkit();

    /**
     * An action to open the options frame.
     */
    public static final AbstractAction2 OPTIONS_ACTION = new OptionsAction();

    public static final int FILE_HISTORY_SIZE = 6;

    /**
     * The actions to load a given recently opened file.
     */
    public static final AbstractAction2 FILE_HISTORY_ACTIONS[]
        = new AbstractAction2[FILE_HISTORY_SIZE];

    public static final String FILE_HISTORY_KEYS[] =
    {
        "toy.autogenerate.filesystem#fileHistory0",
        "toy.autogenerate.filesystem#fileHistory1",
        "toy.autogenerate.filesystem#fileHistory2",
        "toy.autogenerate.filesystem#fileHistory3",
        "toy.autogenerate.filesystem#fileHistory4",
        "toy.autogenerate.filesystem#fileHistory5"
    };

    public static final String CURRENT_DIRECTORY_KEY =
        "toy.autogenerate.filesystem#currentDirectory";

    public static final String LAST_FRAME_BOUNDS_KEY =
        "toy.autogenerate.toyFrame#bounds";

    public static final String LAST_VERTICAL_DIVIDER_LOCATION_KEY =
        "toy.autogenerate.toyFrame#verticalDividerLocation";

    public static final String LAST_EDIT_HORIZONTAL_DIVIDER_LOCATION_KEY =
        "toy.autogenerate.toyFrame.editPane#horizontalDividerLocation";

    public static final String LAST_SIM_HORIZONTAL_DIVIDER_LOCATION_KEY =
        "toy.autogenerate.toyFrame.simPane#horizontalDividerLocation";

    public static final String LAST_OPTIONS_FRAME_POSITION_KEY =
        "toy.autogenerate.optionsFrame#position";

    private static final Object NULL_OBJECT = new Object();
    private static final Hashtable DEFAULT_OPTIONS = new Hashtable();

    private static final Dimension INITIAL_SCREEN_SIZE = TOOLKIT.getScreenSize();

    /**
     * Initialize the tree structure and the default options.
     */
    static
    {
        for (int ctr = 0; ctr < FILE_HISTORY_SIZE; ctr++)
            FILE_HISTORY_ACTIONS[ctr] = new FileHistoryAction(ctr);

        ROOT_NODE.add(EDITING_NODE);
        EDITING_NODE.add(EDITING_FONTS_NODE);
        EDITING_NODE.add(EDITING_COLORS_NODE);
        EDITING_NODE.add(EDITING_HIGHLIGHTING_NODE);
        EDITING_NODE.add(EDITING_AUTO_COMPLETE_NODE);
        EDITING_NODE.add(EDITING_CHECKING_NODE);

        ROOT_NODE.add(EXECUTION_NODE);
        EXECUTION_NODE.add(EXECUTION_EXCEPTION_NODE);
        EXECUTION_NODE.add(EXECUTION_PERFORMANCE_NODE);

        DEFAULT_OPTIONS.put(
            FONT_AREA_KEYS[PROGRAM_AREA_INDEX],
            new Font("Monospaced", Font.PLAIN, 12)
            );
        DEFAULT_OPTIONS.put(
            FONT_AREA_KEYS[WARNING_AREA_INDEX],
            new Font("SansSerif", Font.PLAIN, 12)
            );
        DEFAULT_OPTIONS.put(
            FONT_AREA_KEYS[REFERENCE_AREA_INDEX],
            new Font("Monospaced", Font.PLAIN, 12)
            );
        DEFAULT_OPTIONS.put(
            FONT_AREA_KEYS[STDIN_AREA_INDEX],
            new Font("Monospaced", Font.PLAIN, 12)
            );
        DEFAULT_OPTIONS.put(
            FONT_AREA_KEYS[STDOUT_AREA_INDEX],
            new Font("Monospaced", Font.PLAIN, 12)
            );
        DEFAULT_OPTIONS.put(
            FONT_AREA_KEYS[CORE_AREA_INDEX],
            new Font("Monospaced", Font.PLAIN, 12)
            );

        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[FOREGROUND_COLOR_INDEX],
            Color.black
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[FOREGROUND_COLOR_INDEX],
            Boolean.TRUE
            );
        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[BACKGROUND_COLOR_INDEX],
            Color.white
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[BACKGROUND_COLOR_INDEX],
            Boolean.TRUE
            );
        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[CARET_COLOR_INDEX],
            Color.black
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[CARET_COLOR_INDEX],
            Boolean.TRUE
            );
        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[SELECTION_COLOR_INDEX],
            new Color(204, 204, 255)
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[SELECTION_COLOR_INDEX],
            Boolean.FALSE
            );
        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[SELECTED_TEXT_COLOR_INDEX],
            Color.black
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[SELECTED_TEXT_COLOR_INDEX],
            Boolean.TRUE
            );
        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[HIGHLIGHTED_LINE_COLOR_INDEX],
            new Color(186, 186, 255)
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[HIGHLIGHTED_LINE_COLOR_INDEX],
            Boolean.FALSE
            );
        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[HIGHLIGHTED_TEXT_COLOR_INDEX],
            Color.black
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[HIGHLIGHTED_TEXT_COLOR_INDEX],
            Boolean.TRUE
            );
        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[LINE_NUMBERS_COLOR_INDEX],
            new Color(148, 148, 148)
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[LINE_NUMBERS_COLOR_INDEX],
            Boolean.FALSE
            );
        DEFAULT_OPTIONS.put(
            COLOR_COLOR_KEYS[COLUMN_MARKERS_COLOR_INDEX],
            new Color(206, 206, 206)
            );
        DEFAULT_OPTIONS.put(
            COLOR_DISABLED_KEYS[COLUMN_MARKERS_COLOR_INDEX],
            Boolean.FALSE
            );

        DEFAULT_OPTIONS.put(
            HIGHLIGHT_STYLE_KEYS[TProgramDocument.STYLE_IDENTIFIER],
            new HighlightStyle(Color.black, false, false)
            );
        DEFAULT_OPTIONS.put(
            HIGHLIGHT_STYLE_KEYS[TProgramDocument.STYLE_INSTRUCTION],
            new HighlightStyle(Color.blue, false, false)
            );
        DEFAULT_OPTIONS.put(
            HIGHLIGHT_STYLE_KEYS[TProgramDocument.STYLE_KEYWORD],
            new HighlightStyle(Color.black, true, false)
            );
        DEFAULT_OPTIONS.put(
            HIGHLIGHT_STYLE_KEYS[TProgramDocument.STYLE_COMMENT],
            new HighlightStyle(new Color(0, 128, 0), false, false)
            );
        DEFAULT_OPTIONS.put(
            HIGHLIGHT_STYLE_KEYS[TProgramDocument.STYLE_AUTO_COMMENT],
            new HighlightStyle(Color.black, false, false)
            );
        DEFAULT_OPTIONS.put(
            HIGHLIGHT_STYLE_KEYS[TProgramDocument.STYLE_OUTSIDE_MARGIN],
            new HighlightStyle(Color.red, false, false)
            );

        DEFAULT_OPTIONS.put(
            AUTO_COMPLETER_KEY,
            AUTO_COMPLETERS[1]
            );

        DEFAULT_OPTIONS.put(
            IGNORE_WARNINGS_KEY,
            Boolean.FALSE
            );

        DEFAULT_OPTIONS.put(
            AUTO_CHECK_KEY,
            Boolean.FALSE
            );

        for (int ctr = 0; ctr < EXCEPTION_TYPE_KEYS.length; ctr++)
            DEFAULT_OPTIONS.put(
                EXCEPTION_TYPE_KEYS[ctr],
                Boolean.TRUE
                );

        DEFAULT_OPTIONS.put(CLOCK_PERIOD_KEY, new IntStruct(CLOCK_PERIODS[20]));
        DEFAULT_OPTIONS.put(REFRESH_RATE_KEY, new IntStruct(REFRESH_RATES[0]));

        DEFAULT_OPTIONS.put(
            LAST_FRAME_BOUNDS_KEY,
            new Rectangle(
                INITIAL_SCREEN_SIZE.width / 20,
                INITIAL_SCREEN_SIZE.height / 20,
                (INITIAL_SCREEN_SIZE.width * 9) / 10,
                (INITIAL_SCREEN_SIZE.height * 9) / 10
                )
            );

        DEFAULT_OPTIONS.put(
            LAST_VERTICAL_DIVIDER_LOCATION_KEY,
            new IntStruct(-1)
            );

        DEFAULT_OPTIONS.put(
            LAST_EDIT_HORIZONTAL_DIVIDER_LOCATION_KEY,
            new IntStruct(-1)
            );

        DEFAULT_OPTIONS.put(
            LAST_SIM_HORIZONTAL_DIVIDER_LOCATION_KEY,
            new IntStruct(-1)
            );

        DEFAULT_OPTIONS.put(
            LAST_OPTIONS_FRAME_POSITION_KEY,
            new Point(
                (INITIAL_SCREEN_SIZE.width - WIDTH) / 2,
                (INITIAL_SCREEN_SIZE.height - HEIGHT) / 2
                )
            );
    }

    private static final TOptionsFrame INSTANCE = new TOptionsFrame();

    private File currentDirectory;
    private File fileHistory[];
    private boolean ignoreWarnings, autoCheck;
    private IntStruct clockPeriod;
    private Document clockPeriodDocument;
    private IntStruct refreshRate;
    private Document refreshRateDocument;

    private TDefaultExceptionHandler exceptionHandler;

    private PTree tree;

    private JPanel cardPanel;
    private CardLayout cardLayout;

    private TFontChooserPane editingFontsPane;
    private TColorChooserPane editingColorsPane;
    private TStyleChooserPane editingHighlightingPane;
    private TAutoCompleteChooserPane editingAutoCompletePane;
    private TCheckingChooserPane editingCheckingPane;

    private TExceptionChooserPane executionExceptionPane;
    private TPerformanceChooserPane executionPerformancePane;

    private Hashtable appliedOptions;
    private Vector areaComponents[];
    private Vector highlightedTextAreas;

    private Rectangle lastFrameBounds;
    private IntStruct lastVerticalDividerLocation;
    private IntStruct lastEditHorizontalDividerLocation;
    private IntStruct lastSimHorizontalDividerLocation;
    private Point lastOptionsFramePosition;

    /**
     * Creates a new TOptionsFrame.
     */
    private TOptionsFrame()
    {
        super("Visual X-TOY Options");
        setIconImage(TMain.getFrameIcon());

        clockPeriodDocument = new PlainDocument();
        refreshRateDocument = new PlainDocument();

        exceptionHandler = new TDefaultExceptionHandler();
        appliedOptions = new Hashtable();
        areaComponents = new Vector[FONT_AREA_NAMES.length];
        for (int ctr = 0; ctr < FONT_AREA_NAMES.length; ctr++)
            areaComponents[ctr] = new Vector();
        highlightedTextAreas = new Vector();

        // Load the autogenerate options from the TConfigurationManager, using defaults when a
        // stated preference is unavailable or invalid

        {   // Parse the current directory
            String tempString = TConfigurationManager.getProperty(CURRENT_DIRECTORY_KEY);
            
            if (tempString != null)
            {
                File tempFile = new File(tempString);

                if (tempFile.isDirectory() && tempFile.canRead())
                {
                    currentDirectory = tempFile;
                }
            }
        }

        {   // Parse the file history
            fileHistory = new File[FILE_HISTORY_SIZE];
            int historyCtr = 0;

            for (int ctr = 0; ctr < FILE_HISTORY_SIZE; ctr++)
            {
                String tempString = TConfigurationManager.getProperty(FILE_HISTORY_KEYS[ctr]);

                if (tempString != null)
                {
                    File tempFile = new File(tempString);

                    if (tempFile.isFile() && tempFile.canRead())
                    {
                        fileHistory[historyCtr] = tempFile;
                        FILE_HISTORY_ACTIONS[historyCtr].setEnabled(true);
                        FILE_HISTORY_ACTIONS[historyCtr].putValue(Action.NAME, tempFile.getName());
                        historyCtr++;
                    }
                }
            }
        }

        {   // Parse the frame bounds
            String tempString = TConfigurationManager.getProperty(LAST_FRAME_BOUNDS_KEY);

            if (tempString != null && tempString.startsWith("(") && tempString.endsWith(")"))
            {
                int commaIndex1 = tempString.indexOf(",");
                int commaIndex2 = tempString.indexOf(",", commaIndex1 + 1);
                int commaIndex3 = tempString.indexOf(",", commaIndex2 + 2);

                if (commaIndex1 != -1 && commaIndex2 != -1 && commaIndex3 != -1)
                {
                    try
                    {
                        lastFrameBounds = new Rectangle(
                            Integer.parseInt(tempString.substring(1, commaIndex1)),
                            Integer.parseInt(tempString.substring(commaIndex1 + 1, commaIndex2)),
                            Integer.parseInt(tempString.substring(commaIndex2 + 1, commaIndex3)),
                            Integer.parseInt(
                                tempString.substring(commaIndex3 + 1, tempString.length() - 1)
                                )
                            );

                        if (lastFrameBounds.width > INITIAL_SCREEN_SIZE.width)
                            lastFrameBounds.width = INITIAL_SCREEN_SIZE.width;
                        if (lastFrameBounds.width < 112)
                            lastFrameBounds.width = 112;

                        if (lastFrameBounds.height > INITIAL_SCREEN_SIZE.height)
                            lastFrameBounds.height = INITIAL_SCREEN_SIZE.height;
                        if (lastFrameBounds.height < 27)
                            lastFrameBounds.height = 27;

                        if (lastFrameBounds.x < 0)
                            lastFrameBounds.x = 0;
                        if (lastFrameBounds.x + lastFrameBounds.width > INITIAL_SCREEN_SIZE.width)
                            lastFrameBounds.x = INITIAL_SCREEN_SIZE.width - lastFrameBounds.width;

                        if (lastFrameBounds.y < 0)
                            lastFrameBounds.y = 0;
                        if (lastFrameBounds.y + lastFrameBounds.height > INITIAL_SCREEN_SIZE.height)
                            lastFrameBounds.y = INITIAL_SCREEN_SIZE.height - lastFrameBounds.height;
                    }
                    catch (Exception e)
                    {
                        e.printStackTrace();
                    }
                }
            }
            
            if (lastFrameBounds == null)
                lastFrameBounds = new Rectangle(
                    (Rectangle)DEFAULT_OPTIONS.get(LAST_FRAME_BOUNDS_KEY)
                    );

            appliedOptions.put(
                LAST_FRAME_BOUNDS_KEY,
                lastFrameBounds
                );
        }

        {   // Parse the divider locations
            String tempString;

            tempString = TConfigurationManager.getProperty(LAST_VERTICAL_DIVIDER_LOCATION_KEY);

            if (tempString != null)
            {
                try
                {
                    lastVerticalDividerLocation = new IntStruct(Integer.parseInt(tempString));
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }

            if (lastVerticalDividerLocation == null)
                lastVerticalDividerLocation = new IntStruct(
                    (IntStruct)DEFAULT_OPTIONS.get(LAST_VERTICAL_DIVIDER_LOCATION_KEY)
                    );

            appliedOptions.put(
                LAST_VERTICAL_DIVIDER_LOCATION_KEY,
                lastVerticalDividerLocation
                );


            tempString = TConfigurationManager.getProperty(
                LAST_EDIT_HORIZONTAL_DIVIDER_LOCATION_KEY
                );

            if (tempString != null)
            {
                try
                {
                    lastEditHorizontalDividerLocation = new IntStruct(Integer.parseInt(tempString));
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }

            if (lastEditHorizontalDividerLocation == null)
                lastEditHorizontalDividerLocation = new IntStruct(
                    (IntStruct)DEFAULT_OPTIONS.get(LAST_EDIT_HORIZONTAL_DIVIDER_LOCATION_KEY)
                    );

            appliedOptions.put(
                LAST_EDIT_HORIZONTAL_DIVIDER_LOCATION_KEY,
                lastEditHorizontalDividerLocation
                );


            tempString = TConfigurationManager.getProperty(
                LAST_SIM_HORIZONTAL_DIVIDER_LOCATION_KEY
                );

            if (tempString != null)
            {
                try
                {
                    lastSimHorizontalDividerLocation = new IntStruct(Integer.parseInt(tempString));
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }

            if (lastSimHorizontalDividerLocation == null)
                lastSimHorizontalDividerLocation = new IntStruct(
                    (IntStruct)DEFAULT_OPTIONS.get(LAST_SIM_HORIZONTAL_DIVIDER_LOCATION_KEY)
                    );

            appliedOptions.put(
                LAST_SIM_HORIZONTAL_DIVIDER_LOCATION_KEY,
                lastSimHorizontalDividerLocation
                );
        }

        {   // Parse the option frame position
            String tempString = TConfigurationManager.getProperty(LAST_OPTIONS_FRAME_POSITION_KEY);

            if (tempString != null && tempString.startsWith("(") && tempString.endsWith(")"))
            {
                int commaIndex = tempString.indexOf(",");

                if (commaIndex != -1)
                {
                    try
                    {
                        lastOptionsFramePosition = new Point(
                            Integer.parseInt(tempString.substring(1, commaIndex)),
                            Integer.parseInt(
                                tempString.substring(commaIndex + 1, tempString.length() - 1)
                                )
                            );

                        if (lastOptionsFramePosition.x < 0)
                            lastOptionsFramePosition.x = 0;
                        if (lastOptionsFramePosition.x + WIDTH > INITIAL_SCREEN_SIZE.width)
                            lastOptionsFramePosition.x = INITIAL_SCREEN_SIZE.width - WIDTH;

                        if (lastOptionsFramePosition.y < 0)
                            lastOptionsFramePosition.y = 0;
                        if (lastOptionsFramePosition.y + HEIGHT > INITIAL_SCREEN_SIZE.height)
                            lastOptionsFramePosition.y = INITIAL_SCREEN_SIZE.height - HEIGHT;
                    }
                    catch (Exception e)
                    {
                        e.printStackTrace();
                    }
                }
            }
            
            if (lastOptionsFramePosition == null)
                lastOptionsFramePosition = new Point(
                    (Point)DEFAULT_OPTIONS.get(LAST_OPTIONS_FRAME_POSITION_KEY)
                    );

            appliedOptions.put(
                LAST_OPTIONS_FRAME_POSITION_KEY,
                lastOptionsFramePosition
                );
        }

        // Load the applied options from the TConfigurationManager, using defaults when a
        // stated preference is unavailable or invalid
        {
            // Parse the fonts
            for (int ctr = 0; ctr < FONT_AREA_NAMES.length; ctr++)
            {
                String tempString = TConfigurationManager.getProperty(FONT_AREA_KEYS[ctr]);
                if (tempString != null)
                {
                    appliedOptions.put(FONT_AREA_KEYS[ctr], Font.decode(tempString));
                }
                else {
                    appliedOptions.put(
                        FONT_AREA_KEYS[ctr],
                        DEFAULT_OPTIONS.get(FONT_AREA_KEYS[ctr])
                        );
                }
            }

            // Parse the colors
            for (int ctr = 0; ctr < COLOR_NAMES.length; ctr++)
            {
                String tempString = TConfigurationManager.getProperty(COLOR_COLOR_KEYS[ctr]);
                if (tempString != null && tempString.length() == 7 &&
                    tempString.charAt(0) == '#' &&
                    TWord.isHexDigit(tempString.charAt(1)) &&
                    TWord.isHexDigit(tempString.charAt(2)) &&
                    TWord.isHexDigit(tempString.charAt(3)) &&
                    TWord.isHexDigit(tempString.charAt(4)) &&
                    TWord.isHexDigit(tempString.charAt(5)) &&
                    TWord.isHexDigit(tempString.charAt(6)))
                {
                    appliedOptions.put(
                        COLOR_COLOR_KEYS[ctr],
                        new Color(
                            16 * TWord.hexDigitToInt(tempString.charAt(1)) +
                                 TWord.hexDigitToInt(tempString.charAt(2)),
                            16 * TWord.hexDigitToInt(tempString.charAt(3)) +
                                 TWord.hexDigitToInt(tempString.charAt(4)),
                            16 * TWord.hexDigitToInt(tempString.charAt(5)) +
                                 TWord.hexDigitToInt(tempString.charAt(6))
                            )
                        );
                }
                else {
                    appliedOptions.put(
                        COLOR_COLOR_KEYS[ctr],
                        DEFAULT_OPTIONS.get(COLOR_COLOR_KEYS[ctr])
                        );
                }


                tempString = TConfigurationManager.getProperty(COLOR_DISABLED_KEYS[ctr]);
                if ("true".equals(tempString))
                {
                    appliedOptions.put(COLOR_DISABLED_KEYS[ctr], Boolean.TRUE);
                }
                else if ("false".equals(tempString))
                {
                    appliedOptions.put(COLOR_DISABLED_KEYS[ctr], Boolean.FALSE);
                }
                else {
                    appliedOptions.put(
                        COLOR_DISABLED_KEYS[ctr],
                        DEFAULT_OPTIONS.get(COLOR_DISABLED_KEYS[ctr])
                        );
                }
            }

            // Parse the styles
            for (int ctr = 0; ctr < HIGHLIGHT_STYLE_NAMES.length; ctr++)
            {
                String tempString = TConfigurationManager.getProperty(HIGHLIGHT_STYLE_KEYS[ctr]);

                HighlightStyle style = null;

                if (tempString != null)
                    style = HighlightStyle.parseStyle(tempString);

                if (style == null)
                    style = (HighlightStyle)DEFAULT_OPTIONS.get(HIGHLIGHT_STYLE_KEYS[ctr]);

                appliedOptions.put(HIGHLIGHT_STYLE_KEYS[ctr], style);
            }

            {   // Parse the autoCompleter name
                String tempString = TConfigurationManager.getProperty(AUTO_COMPLETER_KEY);
                AutoCompleter autoCompleter = null;

                for (int ctr = 0; ctr < AUTO_COMPLETER_NAMES.length; ctr++)
                {
                    if (AUTO_COMPLETER_NAMES[ctr].equals(tempString))
                    {
                        autoCompleter = AUTO_COMPLETERS[ctr];
                        break;
                    }
                }

                if (autoCompleter == null)
                    autoCompleter = (AutoCompleter)DEFAULT_OPTIONS.get(AUTO_COMPLETER_KEY);

                appliedOptions.put(AUTO_COMPLETER_KEY, autoCompleter);
            }

            {   // Parse the ignoreWarnings and autoCheck flags
                String tempString = TConfigurationManager.getProperty(IGNORE_WARNINGS_KEY);

                if ("true".equals(tempString))
                {
                    appliedOptions.put(IGNORE_WARNINGS_KEY, Boolean.TRUE);
                }
                else if ("false".equals(tempString))
                {
                    appliedOptions.put(IGNORE_WARNINGS_KEY, Boolean.FALSE);
                }
                else {
                    appliedOptions.put(
                        IGNORE_WARNINGS_KEY,
                        DEFAULT_OPTIONS.get(IGNORE_WARNINGS_KEY)
                        );
                }

                tempString = TConfigurationManager.getProperty(AUTO_CHECK_KEY);

                if ("true".equals(tempString))
                {
                    appliedOptions.put(AUTO_CHECK_KEY, Boolean.TRUE);
                }
                else if ("false".equals(tempString))
                {
                    appliedOptions.put(AUTO_CHECK_KEY, Boolean.FALSE);
                }
                else {
                    appliedOptions.put(
                        AUTO_CHECK_KEY,
                        DEFAULT_OPTIONS.get(AUTO_CHECK_KEY)
                        );
                }
            }

            // Parse the exception type will-throw flags
            for (int ctr = 0; ctr < EXCEPTION_TYPE_KEYS.length; ctr++)
            {
                String tempString = TConfigurationManager.getProperty(EXCEPTION_TYPE_KEYS[ctr]);

                if ("true".equals(tempString))
                {
                    appliedOptions.put(EXCEPTION_TYPE_KEYS[ctr], Boolean.TRUE);
                }
                else if ("false".equals(tempString))
                {
                    appliedOptions.put(EXCEPTION_TYPE_KEYS[ctr], Boolean.FALSE);
                }
                else {
                    appliedOptions.put(
                        EXCEPTION_TYPE_KEYS[ctr],
                        DEFAULT_OPTIONS.get(EXCEPTION_TYPE_KEYS[ctr])
                        );
                }
            }

            // Parse the clock period and refresh rate
            {
                String tempString = TConfigurationManager.getProperty(CLOCK_PERIOD_KEY);

                if (tempString != null)
                {
                    try
                    {
                        clockPeriod = new IntStruct(Integer.parseInt(tempString));
                    }
                    catch (NumberFormatException e)
                    {
                        e.printStackTrace();
                    }
                }

                if (clockPeriod == null)
                    clockPeriod = new IntStruct((IntStruct)DEFAULT_OPTIONS.get(CLOCK_PERIOD_KEY));

                appliedOptions.put(CLOCK_PERIOD_KEY, clockPeriod);


                tempString = TConfigurationManager.getProperty(REFRESH_RATE_KEY);

                if (tempString != null)
                {
                    try
                    {
                        refreshRate = new IntStruct(Integer.parseInt(tempString));
                    }
                    catch (NumberFormatException e)
                    {
                        e.printStackTrace();
                    }
                }

                if (refreshRate == null)
                    refreshRate = new IntStruct((IntStruct)DEFAULT_OPTIONS.get(REFRESH_RATE_KEY));

                appliedOptions.put(REFRESH_RATE_KEY, refreshRate);
            }
        }

        // Prepare the menu
        JMenuBar menuBar;
        JMenu menu;
        JMenuItem menuItem;

        // Create the menu bar.
        menuBar = new JMenuBar();
        setJMenuBar(menuBar);

        {
            // Build the edit menu
            menu = new JMenu("Edit");
            menu.setMnemonic(KeyEvent.VK_E);
            menuBar.add(menu);

            menu.add(ClipboardTargetManager.CUT_ACTION.createMenuItem());
            menu.add(ClipboardTargetManager.COPY_ACTION.createMenuItem());
            menu.add(ClipboardTargetManager.PASTE_ACTION.createMenuItem());

            menu.addSeparator();

            menu.add(ClipboardTargetManager.SELECT_ALL_ACTION.createMenuItem());
        }

        getContentPane().setLayout(new GridBagLayout());

        {   // Prepare the tree
            tree = new PTree(new DefaultTreeModel(ROOT_NODE, true));

            tree.setShowsRootHandles(true);
            tree.putClientProperty("PTree.lineStyle", "Angled");
            tree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
            tree.addTreeSelectionListener(this);
            
            for (int ctr = 0; ctr < tree.getRowCount(); ctr++)
                tree.expandRow(ctr);

            getContentPane().add(
                new JScrollPane(
                    tree,
                    JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                    JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED
                    ),
                new GridBagConstraints(
                    0, 0, 1, 1,
                    0.0, 1.0,
                    GridBagConstraints.NORTHWEST,
                    GridBagConstraints.BOTH,
                    new Insets(2, 2, 2, 2),
                    tree.getPreferredSize().width, 0
                    )
                );
        }

        {   // Prepare the card panel
            cardLayout = new CardLayout();
            cardPanel = new JPanel(cardLayout);

            {   // Prepare the root pane
                JPanel rootPane = new JPanel(new GridBagLayout());
                JTextArea textArea = new JTextArea();

                textArea.setLineWrap(true);
                textArea.setWrapStyleWord(true);
                textArea.setEnabled(false);
                textArea.setBackground(null);
                textArea.setDisabledTextColor(Color.black);
                textArea.setText(
                    "Visual X-TOY's Options Dialog allows users to edit its customizable " +
                    "features.  Changes to these preferences will be saved to the toy.conf " +
                    "file in the (located in the user's home directory) upon exit.\n" +
                    "\n" +
                    "In this dialog you can configure settings ranging from the syntax " +
                    "highlighting styles to the kinds of exceptions that will be thrown in " +
                    "debug mode.\n" +
                    "\n" +
                    "Select the subfolders for more detailed information about the options " +
                    "which can be configured here, or go straight to exploring those options."
                    );
                textArea.setMargin(new Insets(2, 2, 2, 2));
                textArea.setBorder(new EtchedBorder());

                rootPane.add(
                    textArea,
                    new GridBagConstraints(
                        0, 0, 1, 1,
                        1.0, 1.0,
                        GridBagConstraints.CENTER,
                        GridBagConstraints.HORIZONTAL,
                        new Insets(2, 50, 2, 50),
                        0, 0
                        )
                    );

                cardPanel.add(rootPane, ROOT_NODE.toString());
            }

            {   // Prepare the editingPane
                JPanel editingPane = new JPanel(new GridBagLayout());
                JTextArea textArea = new JTextArea();

                textArea.setLineWrap(true);
                textArea.setWrapStyleWord(true);
                textArea.setEnabled(false);
                textArea.setBackground(null);
                textArea.setDisabledTextColor(Color.black);
                textArea.setText(
                    "The 'Editing' options configure the look and feel of the TOY program text " +
                    "area:\n" +
                    "\n" +
                    "In the 'Fonts' section, you can configure the fonts used in certain parts " +
                    "of Visual X-TOY.\n" +
                    "\n" +
                    "In the 'Colors' section, you can configure the colors of all the objects " +
                    "in the program text area with the exception of the syntactical structures.\n" +
                    "\n" +
                    "The 'Syntax Highlighting' section allows you to pick the colors and styles " +
                    "(eg. bold) of the syntactical structures of the program text area.\n" +
                    "\n" +
                    "Finally, the 'Auto-Complete' section allows you to turn on actions " +
                    "which automatically occur as you type your program."
                    );
                textArea.setMargin(new Insets(2, 2, 2, 2));
                textArea.setBorder(new EtchedBorder());

                editingPane.add(
                    textArea,
                    new GridBagConstraints(
                        0, 0, 1, 1,
                        1.0, 1.0,
                        GridBagConstraints.CENTER,
                        GridBagConstraints.HORIZONTAL,
                        new Insets(2, 50, 2, 50),
                        0, 0
                        )
                    );
                cardPanel.add(editingPane, EDITING_NODE.toString());
            }

            {   // Prepare the editingFontsPane
                editingFontsPane = new TFontChooserPane(FONT_AREA_NAMES);

                cardPanel.add(editingFontsPane, EDITING_FONTS_NODE.toString());
            }

            {   // Prepare the editingColorsPane
                editingColorsPane = new TColorChooserPane(COLOR_NAMES, COLOR_DISABLED_DESCRIPTIONS);

                cardPanel.add(editingColorsPane, EDITING_COLORS_NODE.toString());
            }

            {   // Prepare the editingHighlightingPane
                editingHighlightingPane = new TStyleChooserPane(HIGHLIGHT_STYLE_NAMES);

                cardPanel.add(editingHighlightingPane, EDITING_HIGHLIGHTING_NODE.toString());
            }

            {   // Prepare the editingAutoCompletePane
                editingAutoCompletePane = new TAutoCompleteChooserPane(AUTO_COMPLETER_NAMES);

                cardPanel.add(editingAutoCompletePane, EDITING_AUTO_COMPLETE_NODE.toString());
            }

            {   // Prepare the editingCheckingPane
                editingCheckingPane = new TCheckingChooserPane();

                cardPanel.add(editingCheckingPane, EDITING_CHECKING_NODE.toString());
            }

            {   // Prepare the executionPane
                JPanel executionPane = new JPanel(new GridBagLayout());
                JTextArea textArea = new JTextArea();

                textArea.setLineWrap(true);
                textArea.setWrapStyleWord(true);
                textArea.setEnabled(false);
                textArea.setBackground(null);
                textArea.setDisabledTextColor(Color.black);
                textArea.setText(
                    "The 'Execution' options configure the way toy programs are run in both " +
                    "Debug and Sim mode.\n" +
                    "\n" +
                    "In the 'Exception Handling' section, you can configure the TOY virtual " +
                    "machine to ignore certain exceptional conditions in Debug mode.  In Sim " +
                    "mode, all exceptional conditions are ignored and handled as the actual " +
                    "TOY machine would handle them.\n" +
                    "\n" +
                    "The 'Performance' section allows you to configure how fast the TOY virtual " +
                    "machine will run and how often the window's display is refreshed during " +
                    "execution."
                    );
                textArea.setMargin(new Insets(2, 2, 2, 2));
                textArea.setBorder(new EtchedBorder());

                executionPane.add(
                    textArea,
                    new GridBagConstraints(
                        0, 0, 1, 1,
                        1.0, 1.0,
                        GridBagConstraints.CENTER,
                        GridBagConstraints.HORIZONTAL,
                        new Insets(2, 50, 2, 50),
                        0, 0
                        )
                    );
                cardPanel.add(executionPane, EXECUTION_NODE.toString());
            }

            {   // Prepare the executionExceptionPane
                executionExceptionPane = new TExceptionChooserPane();

                cardPanel.add(executionExceptionPane, EXECUTION_EXCEPTION_NODE.toString());
            }

            {   // Prepare the executionPerformancePane
                executionPerformancePane = new TPerformanceChooserPane(
                    CLOCK_PERIODS,
                    REFRESH_RATES
                    );

                cardPanel.add(executionPerformancePane, EXECUTION_PERFORMANCE_NODE.toString());
            }

            getContentPane().add(
                cardPanel,
                new GridBagConstraints(
                    1, 0, 1, 1,
                    2.0, 1.0,
                    GridBagConstraints.NORTHWEST,
                    GridBagConstraints.BOTH,
                    new Insets(2, 2, 2, 2),
                    0, 100
                    )
                );
        }

        {   // Prepare the button panel
            JPanel buttonPanel = new JPanel(new GridLayout(1, 4, 4, 0));
            JButton button;

            button = new JButton("OK");
            button.setActionCommand(OK_COMMAND);
            button.addActionListener(this);
            buttonPanel.add(button);

            button = new JButton("Cancel");
            button.setActionCommand(CANCEL_COMMAND);
            button.addActionListener(this);
            buttonPanel.add(button);

            button = new JButton("Apply");
            button.setActionCommand(APPLY_COMMAND);
            button.addActionListener(this);
            buttonPanel.add(button);

            button = new JButton("Defaults");
            button.setActionCommand(DEFAULTS_COMMAND);
            button.addActionListener(this);
            buttonPanel.add(button);

            getContentPane().add(
                buttonPanel,
                new GridBagConstraints(
                    0, 1, 2, 1,
                    3.0, 0.0,
                    GridBagConstraints.EAST,
                    GridBagConstraints.NONE,
                    new Insets(2, 2, 2, 2),
                    0, 0
                    )
                );
        }

        setBounds(lastOptionsFramePosition.x, lastOptionsFramePosition.y, WIDTH, HEIGHT);
        setResizable(false);

        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        enableEvents(AWTEvent.COMPONENT_EVENT_MASK | AWTEvent.WINDOW_EVENT_MASK);

        setComponentState(appliedOptions);
        doCommand(APPLY_COMMAND, null);

        Runtime.getRuntime().addShutdownHook(new Thread(new ShutdownHook()));
    }

    /**
     * Override this method to prevent unclean disposals.
     */
    public void dispose()
    {
        doCommand(CANCEL_COMMAND, null);
    }

    /**
     * Shows a particular pane in the TOptionsFrame.  The TOptionsFrame is made visible and brought
     * to the front.
     *
     * @param node The node of the pane which should be shown.  A null value will result in a
     * NullPointerException.
     */
    public static void show(DefaultMutableTreeNode node)
    {
        if (node == null)
        {
            throw new NullPointerException();
        }

        INSTANCE.tree.setSelectionPath(new TreePath(node.getPath()));
        INSTANCE.setVisible(true);
        INSTANCE.toFront();
    }

    /**
     * Registers a component to receive font assignments for a particular area.
     *
     * @param component The component to register.  If this is null, a NullPointerException will
     * be thrown.
     * @param areaIndex The index of the area for which the component will receive font assignments.
     * If this is invalid, an ArrayIndexOutOfBoundsException will be thrown.
     */
    public static void registerComponent(Component comp, int areaIndex)
    {
        if (comp == null)
            throw new NullPointerException();

        INSTANCE.areaComponents[areaIndex].add(comp);

        comp.setFont((Font)INSTANCE.appliedOptions.get(FONT_AREA_KEYS[areaIndex]));
    }

    /**
     * Unregisters a component which previously received font assignments for a particular area.
     *
     * @param component The component to unregister.  If this is null, a NullPointerException will
     * be thrown.
     * @param areaIndex The index of the area for which the component will no longer receive font
     * assignments.  If this is invalid, an ArrayIndexOutOfBoundsException will be thrown.
     */
    public static void unregisterComponent(Component comp, int areaIndex)
    {
        if (comp == null)
            throw new NullPointerException();

        INSTANCE.areaComponents[areaIndex].remove(comp);
    }

    /**
     * Registers a PHighlightedTextArea to receive non-font assignments (color, column markers,
     * etc...).
     *
     * @param textArea The text area to register.  If this is null, a NullPointerException will
     * be thrown.
     */
    public static void registerHighlightedTextArea(PHighlightedTextArea textArea)
    {
        if (textArea == null)
            throw new NullPointerException();

        INSTANCE.highlightedTextAreas.add(textArea);

        Hashtable appliedOptions = INSTANCE.appliedOptions;

        boolean disabled;
        Color color;

        // Apply the foreground color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[FOREGROUND_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[FOREGROUND_COLOR_INDEX]);

        textArea.setForeground(disabled?Color.black:color);

        // Apply the background color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[BACKGROUND_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[BACKGROUND_COLOR_INDEX]);

        textArea.setBackground(disabled?Color.white:color);

        // Apply the caret color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[CARET_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[CARET_COLOR_INDEX]);

        textArea.setCaretColor(disabled?null:color);

        // Apply the selection color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[SELECTION_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[SELECTION_COLOR_INDEX]);

        textArea.setSelectionColor(disabled?null:color);

        // Apply the selected text color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[SELECTED_TEXT_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[SELECTED_TEXT_COLOR_INDEX]);

        textArea.setSelectedTextColor(disabled?null:color);

        // Apply the highlighted line color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[HIGHLIGHTED_LINE_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[HIGHLIGHTED_LINE_COLOR_INDEX]);

        if (!disabled)
        {
            textArea.setHighlightColor(color);
            if (textArea.getHighlightedLine() == PHighlightedTextArea.NO_HIGHLIGHT)
                textArea.setHighlightedLine(PHighlightedTextArea.FOLLOW_CARET);
        }
        else {
            if (textArea.getHighlightedLine() == PHighlightedTextArea.FOLLOW_CARET)
                textArea.setHighlightedLine(PHighlightedTextArea.NO_HIGHLIGHT);
        }

        // Apply the highlighted text color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[HIGHLIGHTED_TEXT_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[HIGHLIGHTED_TEXT_COLOR_INDEX]);

        textArea.setHighlightedTextColor(disabled?null:color);

        // Apply the line numbers color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[LINE_NUMBERS_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[LINE_NUMBERS_COLOR_INDEX]);

        if (!disabled)
        {
            textArea.setShowLineNumbers(true);
            textArea.setLineNumberColor(color);
        }
        else {
            textArea.setShowLineNumbers(false);
        }

        // Apply the column markers color option
        disabled = ((Boolean)appliedOptions.get(
            COLOR_DISABLED_KEYS[COLUMN_MARKERS_COLOR_INDEX])
            ).booleanValue();
        color = (Color)appliedOptions.get(COLOR_COLOR_KEYS[COLUMN_MARKERS_COLOR_INDEX]);

        textArea.setMarkerColor(disabled?null:color);

        // Apply the styles
        for (byte ctr = 0; ctr < HIGHLIGHT_STYLE_NAMES.length; ctr++)
        {
            textArea.setStyle(
                (HighlightStyle)appliedOptions.get(HIGHLIGHT_STYLE_KEYS[ctr]),
                ctr
                );
        }

        // Apply the autocompleter
        textArea.setAutoCompleter((AutoCompleter)appliedOptions.get(AUTO_COMPLETER_KEY));
    }

    /**
     * Unregisters a component which previously received non-font assignments.
     *
     * @param component The text area to unregister.  If this is null, a NullPointerException will
     * be thrown.
     */
    public static void unregisterHighlightedTextArea(PHighlightedTextArea textArea)
    {
        if (textArea == null)
            throw new NullPointerException();

        INSTANCE.highlightedTextAreas.remove(textArea);
    }

    /**
     * Gets the exception handler that all TVirtualMachines should use.
     *
     * @return The TExceptionHandler that all TVirtualMachines should use.  The will-throw flags
     * of the returned instance will be kept in-sync with the options that the user has applied.
     */
    public static TExceptionHandler getExceptionHandler()
    {
        return INSTANCE.exceptionHandler;
    }

    /**
     * Returns the clock period (in milliseconds) that the vitrual machine should attempt to attain.
     *
     * @return The clock period (in milliseconds) that the vitrual machine should attempt to attain.
     */
    public static int getClockPeriod()
    {
        return INSTANCE.clockPeriod.value;
    }

    /**
     * Returns a document containing a text representation of the clock period that the vitrual
     * machine should attempt to attain.  This TOptionsFrame will keep that document updated.
     *
     * @return The a document containing a text representation of the clock period that the vitrual
     * machine should attempt to attain.
     */
    public static Document getClockPeriodDocument()
    {
        return INSTANCE.clockPeriodDocument;
    }

    /**
     * Returns the refresh rate (in steps per refresh) that the TFrame should use in rendering a
     * running toy machine.
     *
     * @return The refresh rate (in steps per refresh) that the TFrame should use in rendering a
     * running toy machine.
     */
    public static int getRefreshRate()
    {
        return INSTANCE.refreshRate.value;
    }

    /**
     * Returns a document containing a text representation of the refresh rate that the TFrame
     * should use in rendering a running toy machine.  This TOptionsFrame will keep that document
     * updated.
     *
     * @return The a document containing a text representation of the refresh rate that the TFrame
     * should use in rendering a running toy machine.
     */
    public static Document getRefreshRateDocument()
    {
        return INSTANCE.refreshRateDocument;
    }

    /**
     * Returns wheter or not regular warnings should be ignored when the syntax of a program
     * is being checked.
     *
     * @return True iff regular warnings should be ignored when the syntax of a program
     * is being checked.
     */
    public static boolean getIgnoreWarnings()
    {
        return INSTANCE.ignoreWarnings;
    }

    /**
     * Returns wheter or not the syntax should be checked automatically as the user types.
     *
     * @return True iff the syntax should be checked automatically as the user types.
     */
    public static boolean getAutoCheck()
    {
        return INSTANCE.autoCheck;
    }

    /**
     * Gets the directory in which JFileChoosers should start.
     *
     * @return The File object associated with the &quot;current directory&quot; of the program.
     * Null is returned JFileChooser's default should be used.
     */
    public static File getCurrentDirectory()
    {
        return INSTANCE.currentDirectory;
    }

    /**
     * Sets the directory in which JFileChoosers should start.
     *
     * @param currentDirectory The File object that should be associated with the &quot;current
     * directory&quot; of the program.  If null is passed, a NullPointerException will be thrown.
     * If the File object does not correspond to an existing directory, the command will be ignored.
     */
    public static void setCurrentDirectory(File currentDirectory)
    {
        if (currentDirectory.canRead() && currentDirectory.isDirectory())
        {
            INSTANCE.currentDirectory = currentDirectory;
        }
    }

    /**
     * Gets a File object from the list of recent files.
     *
     * @param index The index of the requested file. An invalid value will result in an
     * ArrayIndexOutOfBoundsException.
     * @return The file object at the given index.  If no file has been specified for that index,
     * null is returned.
     */
    public static File getFileFromHistory(int index)
    {
        return INSTANCE.fileHistory[index];
    }

    /**
     * Puts a File object in the list of recent files.  If that File object is already in the list,
     * it will be brought to the top.
     *
     * @param file The File object that should be addedIf the File object does not correspond to
     * a readable file, the command will be ignored.
     */
    public static void putFileIntoHistory(File file)
    {
        if (file.canRead() && file.isFile())
        {
            File fileHistory[] = INSTANCE.fileHistory;
            synchronized (fileHistory)
            {
                // Update the file history array
                int oldIndex = FILE_HISTORY_SIZE - 1;

                for (int ctr = 0; ctr < FILE_HISTORY_SIZE; ctr++)
                {
                    if (fileHistory[ctr] == file)
                        oldIndex = ctr;
                }

                for (int ctr = oldIndex; ctr > 0; ctr--)
                    fileHistory[ctr] = fileHistory[ctr - 1];

                fileHistory[0] = file;

                // Update the FileHistoryActions
                for (int ctr = 0; ctr < FILE_HISTORY_SIZE; ctr++)
                {
                    if (fileHistory[ctr] == null)
                    {
                        FILE_HISTORY_ACTIONS[ctr].setEnabled(false);
                        FILE_HISTORY_ACTIONS[ctr].putValue(Action.NAME, "Recent File " + ctr);
                    }
                    else {
                        FILE_HISTORY_ACTIONS[ctr].setEnabled(true);
                        FILE_HISTORY_ACTIONS[ctr].putValue(Action.NAME, fileHistory[ctr].getName());
                    }
                }
            }
        }
    }

    /**
     * Sets the bounds for the last TFrame which was closed.  The fields of the given Rectangle
     * are copied over to the Rectangle maintained internally.
     *
     * @param bounds The bounds of the last TFrame which was closed.
     */
    public static void setLastFrameBounds(Rectangle bounds)
    {
        INSTANCE.lastFrameBounds.x = bounds.x;
        INSTANCE.lastFrameBounds.y = bounds.y;
        INSTANCE.lastFrameBounds.width = bounds.width;
        INSTANCE.lastFrameBounds.height = bounds.height;
    }

    /**
     * Returns a copy of the bounds for the last TFrame which was closed.
     *
     * @return The bounds of the last TFrame which was closed.
     */
    public static Rectangle getLastFrameBounds()
    {
        return new Rectangle(INSTANCE.lastFrameBounds);
    }

    /**
     * Sets the location of the vertical divider of the last TFrame which was closed.
     *
     * @param location The location of the vertical divider of the last TFrame which was closed.
     */
    public static void setLastVerticalDividerLocation(int location)
    {
        INSTANCE.lastVerticalDividerLocation.value = location;
    }

    /**
     * Returns the location of the vertical divider of the last TFrame which was closed.
     *
     * @return The location of the vertical divider of the last TFrame which was closed.
     */
    public static int getLastVerticalDividerLocation()
    {
        return INSTANCE.lastVerticalDividerLocation.value;
    }

    /**
     * Sets the location of the horizontal divider of the edit pane of the last TFrame which was
     * closed.
     *
     * @param location The location of the horizontal divider of the edit pane of the last TFrame
     * which was closed.
     */
    public static void setLastEditHorizontalDividerLocation(int location)
    {
        INSTANCE.lastEditHorizontalDividerLocation.value = location;
    }

    /**
     * Returns the location of the horizontal divider of the edit pane of the last TFrame which was
     * closed.
     *
     * @return The location of the horizontal divider of the edit pane of the last TFrame which was
     * closed.
     */
    public static int getLastEditHorizontalDividerLocation()
    {
        return INSTANCE.lastEditHorizontalDividerLocation.value;
    }

    /**
     * Sets the location of the horizontal divider of the sim pane of the last TFrame which was
     * closed.
     *
     * @param location The location of the horizontal divider of the sim pane of the last TFrame
     * which was closed.
     */
    public static void setLastSimHorizontalDividerLocation(int location)
    {
        INSTANCE.lastSimHorizontalDividerLocation.value = location;
    }

    /**
     * Returns the location of the horizontal divider of the sim pane of the last TFrame which was
     * closed.
     *
     * @return The location of the horizontal divider of the sim pane of the last TFrame which was
     * closed.
     */
    public static int getLastSimHorizontalDividerLocation()
    {
        return INSTANCE.lastSimHorizontalDividerLocation.value;
    }

    /**
     * Sets the state of the components to the values in the given Hashtable.
     *
     * @param options The Hashtable containing all the values to restore.
     */
    private void setComponentState(Hashtable options)
    {
        // Restore all of the fonts
        for (int ctr = 0; ctr < FONT_AREA_NAMES.length; ctr++)
        {
            Object value = options.get(FONT_AREA_KEYS[ctr]);

            if (value != null)
                editingFontsPane.setFont((Font)value, ctr);
        }

        // Restore all of the colors
        for (int ctr = 0; ctr < COLOR_NAMES.length; ctr++)
        {
            Object value = options.get(COLOR_COLOR_KEYS[ctr]);

            if (value != null)
                editingColorsPane.setColor((Color)value, ctr);

            value = options.get(COLOR_DISABLED_KEYS[ctr]);
            if (value != null)
                editingColorsPane.setDisabled(((Boolean)value).booleanValue(), ctr);
        }

        // Restore all of the highlight styles
        for (byte ctr = 0; ctr < HIGHLIGHT_STYLE_NAMES.length; ctr++)
        {
            Object value = options.get(HIGHLIGHT_STYLE_KEYS[ctr]);

            if (value != null)
                editingHighlightingPane.setStyle((HighlightStyle)value, ctr);
        }


        {   // Restore the autocompleter
            Object value = options.get(AUTO_COMPLETER_KEY);

            for (int ctr = 0; ctr < AUTO_COMPLETERS.length; ctr++)
            {
                if (AUTO_COMPLETERS[ctr] == value)
                    editingAutoCompletePane.setAutoCompleterId(ctr);
            }
        }


        {   // Restore all of the syntax checking flags
            Object value = options.get(IGNORE_WARNINGS_KEY);

            if (value != null)
                editingCheckingPane.setIgnoreWarnings(((Boolean)value).booleanValue());

            value = options.get(AUTO_CHECK_KEY);

            if (value != null)
                editingCheckingPane.setAutoCheck(((Boolean)value).booleanValue());
        }

        // Restore all of the exception type ignoring flags
        for (byte ctr = 0; ctr < EXCEPTION_TYPE_KEYS.length; ctr++)
        {
            Object value = options.get(EXCEPTION_TYPE_KEYS[ctr]);

            if (value != null)
                executionExceptionPane.setWillThrow(ctr, ((Boolean)value).booleanValue());
        }

        // Restore the clock period and refresh rates
        {
            Object value = options.get(CLOCK_PERIOD_KEY);

            if (value != null)
                executionPerformancePane.setClockPeriod(((IntStruct)value).value);

            value = options.get(REFRESH_RATE_KEY);

            if (value != null)
                executionPerformancePane.setRefreshRate(((IntStruct)value).value);
        }
    }

    /**
     * Performs a command based on the argument.
     *
     * @param command A string representing the command.  Note that pointer equality (not string
     * equality) is tested here, so it is important to use the string constants defined in this
     * class.  An IllegalArgumentException will be thrown if the argument is invalid.
     * @return True iff the command was executed sucessfully.
     */
    public boolean doCommand(String command, Object extraInfo)
    {
        ////////////////////////////////////////////////////////////////////////////////////////////
        if (command == OK_COMMAND)
        {
            doCommand(APPLY_COMMAND, null);

            setVisible(false);
            return true;
        }
        ////////////////////////////////////////////////////////////////////////////////////////////
        if (command == CANCEL_COMMAND)
        {
            setComponentState(appliedOptions);

            setVisible(false);
            return true;
        }
        ////////////////////////////////////////////////////////////////////////////////////////////
        if (command == APPLY_COMMAND)
        {
            // Move the data from the components to the Hashtable and propagate those changes as
            // necessary
            Object array[];
            int length;

            // Apply all of the fonts
            for (int ctr = 0; ctr < FONT_AREA_NAMES.length; ctr++)
            {
                Font font = editingFontsPane.getFont(ctr);

                appliedOptions.put(FONT_AREA_KEYS[ctr], font);

                array = areaComponents[ctr].toArray();
                length = array.length;

                for (int ctr2 = 0; ctr2 < length; ctr2++)
                    ((Component)array[ctr2]).setFont(font);
            }

            // Apply all of the colors
            array = highlightedTextAreas.toArray();
            length = array.length;

            for (int ctr = 0; ctr < COLOR_NAMES.length; ctr++)
            {
                Color color = editingColorsPane.getColor(ctr);
                boolean disabled = editingColorsPane.isDisabled(ctr);

                appliedOptions.put(COLOR_COLOR_KEYS[ctr], color);
                appliedOptions.put(COLOR_DISABLED_KEYS[ctr], disabled?Boolean.TRUE:Boolean.FALSE);

                // Act based on which color we're talking about
                switch (ctr)
                {
                  case FOREGROUND_COLOR_INDEX:
                    for (int ctr2 = 0; ctr2 < length; ctr2++)
                    {
                        ((PHighlightedTextArea)array[ctr2]).setForeground(
                            disabled?Color.black:color
                            );
                    }
                    break;

                  case BACKGROUND_COLOR_INDEX:
                    for (int ctr2 = 0; ctr2 < length; ctr2++)
                    {
                        ((PHighlightedTextArea)array[ctr2]).setBackground(
                            disabled?Color.white:color
                            );
                    }
                    break;

                  case CARET_COLOR_INDEX:
                    for (int ctr2 = 0; ctr2 < length; ctr2++)
                    {
                        ((PHighlightedTextArea)array[ctr2]).setCaretColor(
                            disabled?null:color
                            );
                    }
                    break;

                  case SELECTION_COLOR_INDEX:
                    for (int ctr2 = 0; ctr2 < length; ctr2++)
                    {
                        ((PHighlightedTextArea)array[ctr2]).setSelectionColor(
                            disabled?null:color
                            );
                    }
                    break;

                  case SELECTED_TEXT_COLOR_INDEX:
                    for (int ctr2 = 0; ctr2 < length; ctr2++)
                    {
                        ((PHighlightedTextArea)array[ctr2]).setSelectedTextColor(
                            disabled?null:color
                            );
                    }
                    break;

                  case HIGHLIGHTED_LINE_COLOR_INDEX:
                    if (!disabled)
                    {
                        for (int ctr2 = 0; ctr2 < length; ctr2++)
                        {
                            PHighlightedTextArea textArea = ((PHighlightedTextArea)array[ctr2]);
                            textArea.setHighlightColor(color);
                            if (textArea.getHighlightedLine() == PHighlightedTextArea.NO_HIGHLIGHT)
                                textArea.setHighlightedLine(PHighlightedTextArea.FOLLOW_CARET);
                        }
                    }
                    else {
                        for (int ctr2 = 0; ctr2 < length; ctr2++)
                        {
                            PHighlightedTextArea textArea = ((PHighlightedTextArea)array[ctr2]);
                            if (textArea.getHighlightedLine() == PHighlightedTextArea.FOLLOW_CARET)
                                textArea.setHighlightedLine(PHighlightedTextArea.NO_HIGHLIGHT);
                        }
                    }
                    break;

                  case HIGHLIGHTED_TEXT_COLOR_INDEX:
                    for (int ctr2 = 0; ctr2 < length; ctr2++)
                    {
                        ((PHighlightedTextArea)array[ctr2]).setHighlightedTextColor(
                            disabled?null:color
                            );
                    }
                    break;

                  case LINE_NUMBERS_COLOR_INDEX:
                    if (!disabled)
                    {
                        for (int ctr2 = 0; ctr2 < length; ctr2++)
                        {
                            PHighlightedTextArea textArea = ((PHighlightedTextArea)array[ctr2]);
                            textArea.setShowLineNumbers(true);
                            textArea.setLineNumberColor(color);
                        }
                    }
                    else {
                        for (int ctr2 = 0; ctr2 < length; ctr2++)
                            ((PHighlightedTextArea)array[ctr2]).setShowLineNumbers(false);
                    }
                    break;

                  case COLUMN_MARKERS_COLOR_INDEX:
                    for (int ctr2 = 0; ctr2 < length; ctr2++)
                    {
                        ((PHighlightedTextArea)array[ctr2]).setMarkerColor(
                            disabled?null:color
                            );
                    }
                    break;

                  default:
                    throw new RuntimeException();
                }
            }

            // Apply the styles
            for (byte ctr = 0; ctr < HIGHLIGHT_STYLE_NAMES.length; ctr++)
            {
                HighlightStyle style = editingHighlightingPane.getStyle(ctr);

                appliedOptions.put(HIGHLIGHT_STYLE_KEYS[ctr], style);

                for (int ctr2 = 0; ctr2 < length; ctr2++)
                {
                    ((PHighlightedTextArea)array[ctr2]).setStyle(style, ctr);
                }
            }

            {   // Apply the autocompleter
                int autoCompleterId = editingAutoCompletePane.getAutoCompleterId();

                AutoCompleter autoCompleter = AUTO_COMPLETERS[autoCompleterId];
                appliedOptions.put(AUTO_COMPLETER_KEY, autoCompleter);

                for (int ctr2 = 0; ctr2 < length; ctr2++)
                {
                    ((PHighlightedTextArea)array[ctr2]).setAutoCompleter(autoCompleter);
                }
            }

            {   // Apply the syntax checking flags
                ignoreWarnings = editingCheckingPane.getIgnoreWarnings();

                appliedOptions.put(IGNORE_WARNINGS_KEY, ignoreWarnings?Boolean.TRUE:Boolean.FALSE);

                autoCheck = editingCheckingPane.getAutoCheck();

                TFrame.setAutoCheck(autoCheck);
                appliedOptions.put(AUTO_CHECK_KEY, autoCheck?Boolean.TRUE:Boolean.FALSE);
            }

            // Apply the exception type ignoring flags
            for (byte ctr = 0; ctr < EXCEPTION_TYPE_KEYS.length; ctr++)
            {
                boolean willThrow = executionExceptionPane.getWillThrow(ctr);

                appliedOptions.put(EXCEPTION_TYPE_KEYS[ctr], willThrow?Boolean.TRUE:Boolean.FALSE);

                exceptionHandler.setWillThrow(ctr, willThrow);
            }

            clockPeriod.value = executionPerformancePane.getClockPeriod();
            try
            {
                clockPeriodDocument.remove(0, clockPeriodDocument.getLength());
                clockPeriodDocument.insertString(0, clockPeriod + " ms", null);
            }
            catch (BadLocationException e)
            {
                e.printStackTrace();
            }

            refreshRate.value = executionPerformancePane.getRefreshRate();
            try
            {
                refreshRateDocument.remove(0, refreshRateDocument.getLength());
                refreshRateDocument.insertString(0, refreshRate + " spr", null);
            }
            catch (BadLocationException e)
            {
                e.printStackTrace();
            }

            return true;
        }
        ////////////////////////////////////////////////////////////////////////////////////////////
        if (command == DEFAULTS_COMMAND)
        {
            setComponentState(DEFAULT_OPTIONS);

            return true;
        }

        throw new IllegalArgumentException();
    }

    /**
     * Implement ActionListener to pass the command onto doCommand().
     *
     * @see #doCommand(String, Object)
     */
    public void actionPerformed(ActionEvent e)
    {
        doCommand(e.getActionCommand(), null);
    }

    /**
     * Implement TreeSelectionListener to update the cardPanel's selection when the tree's
     * selection changes.
     */
    public void valueChanged(TreeSelectionEvent e)
    {
        TreePath path = e.getPath();

        if (path != null)
        {
            cardLayout.show(cardPanel, path.getLastPathComponent().toString());
        }
    }

    /**
     * Intercept ComponentEvents to pick up COMPONENT_MOVED events.
     */
    protected void processComponentEvent(ComponentEvent e)
    {
        if (e.getID() == ComponentEvent.COMPONENT_MOVED)
        {
            lastOptionsFramePosition.x = getX();
            lastOptionsFramePosition.y = getY();
        }

        super.processComponentEvent(e);
    }

    /**
     * Intercept WindowEvents to pick up WINDOW_CLOSING events.
     */
    protected void processWindowEvent(WindowEvent e)
    {
        if (e.getID() == WindowEvent.WINDOW_CLOSING)
        {
            doCommand(CANCEL_COMMAND, null);
        }

        super.processWindowEvent(e);
    }

    /**
     * IntStruct is a simple structure that encapsulates an integer.
     *
     * @author btsang
     * @version 7.1
     */
    protected static class IntStruct
    {
        protected int value;

        /**
         * Instantiates a new IntStruct.
         */
        protected IntStruct()
        {
            this(0);
        }

        /**
         * Instantiates a new IntStruct.
         */
        protected IntStruct(IntStruct intStruct)
        {
            this(intStruct.value);
        }

        /**
         * Instantiates a new IntStruct.
         */
        protected IntStruct(int value)
        {
            this.value = value;
        }

        /**
         * Override toString to output the value of the IntStruct.
         */
        public String toString()
        {
            return String.valueOf(value);
        }
    }

    /**
     * OptionsAction is an action for showing the TOptionFrame instance.
     *
     * @author btsang
     * @version 7.1
     */
    protected static class OptionsAction extends AbstractAction2
    {
        /**
         * Creates a new OptionsAction.
         */
        protected OptionsAction()
        {
            super("Options...");

            setMnemonic(KeyEvent.VK_O);

            setAccelerator(
                KeyStroke.getKeyStroke(
                    KeyEvent.VK_O,
                    TOOLKIT.getMenuShortcutKeyMask() | ActionEvent.SHIFT_MASK
                    )
                );
        }

        /**
         * Implements ActionListener to perform the redo operation.
         */
        public void actionPerformed(ActionEvent e)
        {
            INSTANCE.setVisible(true);
            INSTANCE.toFront();
        }
    }

    /**
     * FileHistoryAction is an action for opening a file that had been opened in the past.
     *
     * @author btsang
     * @version 7.1
     */
    protected static class FileHistoryAction extends AbstractAction2
    {
        private int index;

        /**
         * Creates a new FileHistoryAction.
         */
        protected FileHistoryAction(int index)
        {
            super("Recent File " + index);

            if (index < 0 || index >= FILE_HISTORY_SIZE)
                throw new ArrayIndexOutOfBoundsException();

            this.index = index;
            setEnabled(false);
        }

        /**
         * Implements ActionListener to perform the redo operation.
         */
        public void actionPerformed(ActionEvent e)
        {
            File file = INSTANCE.fileHistory[index];

            if (file == null)
                return;

            Component source = (Component)e.getSource();

            while (!(source instanceof TFrame))
            {
                if (source instanceof JPopupMenu)
                    source = ((JPopupMenu)source).getInvoker();
                else
                    source = source.getParent();
            }

            ((TFrame)source).doCommand(TFrame.OPEN_COMMAND, file);
        }
    }

    /**
     * ShutdownHook implements Runnable to store the TOptionsFrame's parameters upon exit.
     *
     * @author btsang
     * @version 7.1
     */
    protected class ShutdownHook implements Runnable
    {
        /**
         * Creates a new ShutdownHook.
         */
        protected ShutdownHook()
        {

        }

        /**
         * Saves all the autogenerate fields into a copy of the appliedOptions Mapping and exports
         * it to the toy.conf file.
         */
        public void run()
        {
            Hashtable map = new Hashtable();
            Object keys[] = appliedOptions.keySet().toArray();

            // Transform the values and keys as necessary
            for (int ctr = 0; ctr < keys.length; ctr++)
            {
                Object value = appliedOptions.get(keys[ctr]);

                if (value == NULL_OBJECT)
                {
                    map.put(
                        keys[ctr].toString(),
                        "null"
                        );
                }
                else if (value instanceof Point)
                {
                    Point point = (Point)value;

                    map.put(
                        keys[ctr].toString(),
                        "(" + point.x + "," + point.y + ")"
                        );
                }
                else if (value instanceof Rectangle)
                {
                    Rectangle rectangle = (Rectangle)value;

                    map.put(
                        keys[ctr].toString(),
                        "(" + rectangle.x + "," + rectangle.y + "," + rectangle.width + "," +
                            rectangle.height + ")"
                        );
                }
                else if (value instanceof Font)
                {
                    Font font = (Font)value;

                    map.put(
                        keys[ctr].toString(),
                        font.getName() + "-" +
                            (font.isBold()?"BOLD":"") + (font.isItalic()?"ITALIC":"") + "-" +
                            font.getSize()
                        );
                }
                else if (value instanceof Color)
                {
                    Color color = (Color)value;

                    map.put(
                        keys[ctr].toString(),
                        "#" + TWord.HEX_PAIRS[color.getRed()] + TWord.HEX_PAIRS[color.getGreen()] +
                            TWord.HEX_PAIRS[color.getBlue()]
                        );
                }
                else if (value instanceof AutoCompleter)
                {
                    for (int ctr2 = 0; ctr2 < AUTO_COMPLETERS.length; ctr2++)
                    {
                        if (value == AUTO_COMPLETERS[ctr2])
                        {
                            map.put(
                                keys[ctr].toString(),
                                AUTO_COMPLETER_NAMES[ctr2]
                                );
                            break;
                        }
                    }
                }
                else {
                    map.put(keys[ctr].toString(), value.toString());
                }
            }

            File tempFile = INSTANCE.currentDirectory;

            if (tempFile != null)
            {
                try
                {
                    map.put(CURRENT_DIRECTORY_KEY, tempFile.getCanonicalPath());
                }
                catch(IOException e)
                {
                    e.printStackTrace();
                    map.put(CURRENT_DIRECTORY_KEY, tempFile.getAbsolutePath());
                }
            }

            int historyCtr = 0;
            for (int ctr = 0; ctr < FILE_HISTORY_SIZE; ctr++)
            {
                tempFile = INSTANCE.fileHistory[ctr];
                if (tempFile != null)
                {
                    try
                    {
                        map.put(FILE_HISTORY_KEYS[historyCtr], tempFile.getCanonicalPath());
                        historyCtr++;
                    }
                    catch(IOException e)
                    {
                        e.printStackTrace();
                        map.put(FILE_HISTORY_KEYS[historyCtr], tempFile.getAbsolutePath());
                    }
                }
            }

            TConfigurationManager.exportProperties(map);
        }
    }
}
